package broker_test

import (
	"encoding/json"
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("deprovisioning instances", func() {
	var (
		instanceID = "an-instance-to-be-deprovisioned"

		asyncAllowed    bool
		deprovisionSpec brokerapi.DeprovisionServiceSpec
		deprovisionErr  error

		deleteTaskID = 88
		boshVMs      = bosh.BoshVMs{"instance-group": {"ip.from.bosh"}}
		boshManifest = []byte("name: service-instance_an-instance-to-be-deprovisioned")

		preDeleteDeploymentFeatureFlag bool
	)

	BeforeEach(func() {
		asyncAllowed = true
		serviceAdapter.PreDeleteDeploymentReturns(nil)
		boshClient.VMsReturns(boshVMs, nil)
		boshClient.GetManifestRawReturns(boshManifest, nil)
		boshClient.DeleteDeploymentReturns(deleteTaskID, nil)
		preDeleteDeploymentFeatureFlag = false
	})

	JustBeforeEach(func() {
		b.PreDeleteDeploymentHookFeatureFlag = preDeleteDeploymentFeatureFlag
		deprovisionSpec, deprovisionErr = b.Deprovision(instanceID, brokerapi.DeprovisionDetails{}, asyncAllowed)
	})

	It("returns that is deprovisioning asynchronously", func() {
		Expect(deprovisionSpec.IsAsync).To(BeTrue())
	})

	It("returns the bosh task ID and operation type in the operation data", func() {
		var operationData broker.OperationData
		Expect(json.Unmarshal([]byte(deprovisionSpec.OperationData), &operationData)).To(Succeed())
		Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: deleteTaskID, OperationType: broker.OperationTypeDelete}))
	})

	It("deletes a bosh deployment whose name is based on the instance ID", func() {
		Expect(boshClient.DeleteDeploymentCallCount()).To(Equal(1))
		actualInstanceID, _ := boshClient.DeleteDeploymentArgsForCall(0)
		Expect(actualInstanceID).To(Equal(deploymentName(instanceID)))
	})

	It("logs that it will delete the deployment with a request ID", func() {
		Expect(logBuffer.String()).To(MatchRegexp(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} deleting deployment for instance`))
	})

	It("logs the task id to delete the deployment", func() {
		Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Delete instance %s was %d", instanceID, deleteTaskID)))
	})

	Context("when the async allowed flag is false", func() {
		BeforeEach(func() {
			asyncAllowed = false
		})

		It("returns an error", func() {
			Expect(deprovisionErr).To(Equal(brokerapi.ErrAsyncRequired))
		})
	})

	Describe("the deployment not existing", func() {
		Context("when deleting the deployment", func() {
			BeforeEach(func() {
				boshClient.DeleteDeploymentReturns(0, boshclient.DeploymentNotFoundError{})
			})

			It("returns an error", func() {
				Expect(deprovisionErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error deprovisioning: instance %s, not found", instanceID)))
			})
		})
	})

	Context("when deleting the deployment fails", func() {
		BeforeEach(func() {
			boshClient.DeleteDeploymentReturns(0, errors.New("er ma gerd!"))
		})

		It("returns a generic error message", func() {
			Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("er ma gerd!"))
		})
	})

	// TODO is this feature ever going to be resurrected? If not, delete this flag and any code path that depends on it being true.
	// There is no non-test code path that sets this flag to true, hence all integration tests have been removed.
	Context("whent the pre-delete-deployment feature flag is enabled", func() {
		BeforeEach(func() {
			preDeleteDeploymentFeatureFlag = true
		})

		It("calls the pre-delete-deployment hook on the service adapter", func() {
			Expect(deprovisionErr).NotTo(HaveOccurred())
			Expect(serviceAdapter.PreDeleteDeploymentCallCount()).To(Equal(1))
			actualInstanceID, actualBoshVms, actualManifest, _ := serviceAdapter.PreDeleteDeploymentArgsForCall(0)
			Expect(actualInstanceID).To(Equal(instanceID))
			Expect(actualBoshVms).To(Equal(boshVMs))
			Expect(actualManifest).To(Equal(boshManifest))
		})

		It("fetches VMs for the service instance", func() {
			Expect(boshClient.VMsCallCount()).To(Equal(1))
			actualInstanceID, _ := boshClient.VMsArgsForCall(0)
			Expect(actualInstanceID).To(Equal(deploymentName(instanceID)))
		})

		It("fetches manifest for the service instance", func() {
			Expect(boshClient.GetManifestRawCallCount()).To(Equal(1))
			actualInstanceID, _ := boshClient.GetManifestRawArgsForCall(0)
			Expect(actualInstanceID).To(Equal(deploymentName(instanceID)))
		})

		Context("when the the pre-delete-deployment returns a generic error", func() {
			BeforeEach(func() {
				serviceAdapter.PreDeleteDeploymentReturns(errors.New("complete failure"))
			})

			It("returns an error", func() {
				Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when the the pre-delete-deployment returns an adapter command error", func() {
			BeforeEach(func() {
				serviceAdapter.PreDeleteDeploymentReturns(broker.NewAdapterCommandError("for app dev", "for logging"))
			})

			It("returns an error based on the CLI user component of the adapter error", func() {
				Expect(deprovisionErr).To(MatchError("for app dev"))
			})

			It("logs an error based on the operator component of the adapter error", func() {
				Expect(logBuffer).To(ContainSubstring("for logging"))
			})
		})

		Context("when the adapter has not implemented pre-delete-deployment", func() {
			BeforeEach(func() {
				serviceAdapter.PreDeleteDeploymentReturns(broker.NewServiceAdapterCommandNotImplementedError(fmt.Errorf("not implemented!")))
			})

			It("does not error", func() {
				Expect(deprovisionErr).NotTo(HaveOccurred())
			})

			It("still returns that is deprovisioning asynchronously", func() {
				Expect(deprovisionSpec.IsAsync).To(BeTrue())
			})

			It("still returns the bosh task ID and operation type", func() {
				var operationData broker.OperationData
				Expect(json.Unmarshal([]byte(deprovisionSpec.OperationData), &operationData)).To(Succeed())
				Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: deleteTaskID, OperationType: broker.OperationTypeDelete}))
			})
		})

		Context("when getting vms for the deployment fails", func() {
			BeforeEach(func() {
				boshClient.VMsReturns(bosh.BoshVMs{}, errors.New("er ma gerd!"))
			})

			It("returns a generic error message", func() {
				Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("er ma gerd!"))
			})
		})

		Context("when getting the manifest for the deployment fails", func() {
			BeforeEach(func() {
				boshClient.GetManifestRawReturns([]byte{}, errors.New("er ma gerd!"))
			})

			It("returns a generic error message", func() {
				Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("er ma gerd!"))
			})
		})

		Describe("the deployment not existing", func() {
			Context("when getting vms", func() {
				BeforeEach(func() {
					boshClient.VMsReturns(bosh.BoshVMs{}, boshclient.DeploymentNotFoundError{})
				})

				It("returns an error", func() {
					Expect(deprovisionErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
				})

				It("logs the error", func() {
					Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error deprovisioning: instance %s, not found", instanceID)))
				})
			})

			Context("when getting the manifest", func() {
				BeforeEach(func() {
					boshClient.GetManifestRawReturns([]byte{}, boshclient.DeploymentNotFoundError{})
				})

				It("returns an error", func() {
					Expect(deprovisionErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
				})

				It("logs the error", func() {
					Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error deprovisioning: instance %s, not found", instanceID)))
				})
			})
		})
	})
})
