package broker

import (
	"encoding/json"
	"fmt"
	"log"

	"github.com/pivotal-cf/brokerapi"
)

func (b *Broker) Provision(instanceID string, details brokerapi.ProvisionDetails, asyncAllowed bool) (brokerapi.ProvisionedServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	logger := b.loggerFactory.NewLogger()

	if !asyncAllowed {
		return brokerapi.ProvisionedServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	requestParams, err := convertProvisionDetailsToMap(details)
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	boshTaskID, dashboardURL, err := b.deployInstance(instanceID, details.PlanID, "", requestParams, logger)
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, convertToDisplayableError(err, logger)
	}

	operationData, err := json.Marshal(OperationData{BoshTaskID: boshTaskID, OperationType: OperationTypeCreate})
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	return brokerapi.ProvisionedServiceSpec{IsAsync: true, DashboardURL: dashboardURL, OperationData: string(operationData)}, nil
}

func (b *Broker) deployInstance(instanceID, planID, previousPlanID string, requestParams map[string]interface{}, logger *log.Logger) (int, string, error) {
	errs := func(err error) (int, string, error) {
		logger.Println(err)
		return 0, "", err
	}

	planConfig := b.serviceOffering.FindPlanById(planID)
	if planConfig == nil {
		return errs(DisplayableError{fmt.Errorf("plan %s not found", planID)})
	}

	manifestForDeployment, err := b.boshClient.GetManifestRaw(deploymentName(instanceID), logger)
	if err != nil {
		return errs(fmt.Errorf("error fetching instance: %s", err))
	}

	if manifestForDeployment != nil {
		return errs(DisplayableError{brokerapi.ErrInstanceAlreadyExists})
	}

	if planConfig.Quotas != nil {
		count, countErr := b.cfClient.CountInstancesOfPlan(b.serviceOffering.ID, planID, logger)
		if countErr != nil {
			return errs(fmt.Errorf("error counting instances of plan: %s", countErr))
		}

		if count >= planConfig.Quotas.ServiceInstanceLimit {
			return errs(DisplayableError{brokerapi.ErrPlanQuotaExceeded})
		}
	}

	boshTaskID, manifest, err := b.deployer.Deploy(instanceID, planID, deploymentName(instanceID), OperationTypeCreate, requestParams, nil, nil, logger)
	if err != nil {
		return errs(err)
	}

	dashboardUrl, err := b.serviceAdapter.GenerateDashboardUrl(instanceID, serviceAdapterPlan(*planConfig, b.serviceOffering.GlobalProperties), manifest, logger)
	switch err := err.(type) {
	case nil, ServiceAdapterCommandNotImplementedError:
		return boshTaskID, dashboardUrl, nil
	case error:
		return 0, "", err
	}

	return boshTaskID, dashboardUrl, nil
}
